/*********************************************************************
*                                                                    *
*   This file is made available under the                            *
*   Creative Commons CC0 1.0 Universal Public Domain Dedication.     *
*   (https://creativecommons.org/publicdomain/zero/1.0/legalcode)    *
*                                                                    *
*   The person who associated a work with this deed has dedicated    *
*   the work to the public domain by waiving all of his or her       *
*   rights to the work worldwide under copyright law, including      *
*   all related and neighboring rights, to the extent allowed by     *
*   law.                                                             *
*                                                                    *
*   You can copy, modify, distribute and perform the work, even      *
*   for commercial purposes, all without asking permission.          *
*                                                                    *
**********************************************************************
-------------------------- END-OF-HEADER -----------------------------

File    : main.c
Purpose : Generic application start
*/

/*********************************************************************
*
*       #include Section
*
**********************************************************************
*/
#include <stdio.h>
#include <stdlib.h>

#include "stm32f7xx_hal.h"
#include "stm32f746xx.h"
#include "arm_math.h"
#include "stm32746g_discovery_audio.h"  // <- Modified: DEFAULT_AUDIO_IN_FREQ set to I2S_AUDIOFREQ_48K
#include "stm32746g_discovery_ts.h"
#include "RTOS.h"
#include "GUI.h"
#include "WM.h"
#include "GRAPH.h"
#include "MENU.h"
#include "TEXT.h"


/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

#define FFT_LENGTH	470
#define GRAPH_RANGE_Y 	240
#define GRAPH_OFFSET_Y 	2


/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

#define TASK_EVENT_DMA_HALF_DONE 		0x00000001
#define TASK_EVENT_DMA_DONE 			0x00000002
#define TASK_EVENT_NEW_GESTURE_DATA		0x00000004

#define TASK_EVENT_CHANGE_VIEW_MOVE_LEFT	0x00000100
#define TASK_EVENT_CHANGE_VIEW_MOVE_RIGHT	0x00000200
#define TASK_EVENT_CHANGE_VIEW_MOVE_UP		0x00000400
#define TASK_EVENT_CHANGE_VIEW_MOVE_DOWN	0x00000800
#define TASK_EVENT_CHANGE_VIEW_ZOOM_IN_X	0x00001000
#define TASK_EVENT_CHANGE_VIEW_ZOOM_OUT_X	0x00002000
#define TASK_EVENT_CHANGE_VIEW_ZOOM_IN_Y	0x00004000
#define TASK_EVENT_CHANGE_VIEW_ZOOM_OUT_Y	0x00008000

#define TASK_EVENT_CHANGE_SOURCE_LINE_1         0x00010000
#define TASK_EVENT_CHANGE_SOURCE_LINE_2         0x00020000
#define TASK_EVENT_CHANGE_SOURCE_MIC_1          0x00040000
#define TASK_EVENT_CHANGE_SOURCE_MIC_2          0x00080000

#define DMA_BUFFER_LENGTH			16384
#define SIGNAL_SAMPLES                          (DMA_BUFFER_LENGTH / 4)

#define MENU_ID_LINE_1  GUI_ID_USER + 1
#define MENU_ID_LINE_2  GUI_ID_USER + 2
#define MENU_ID_MIC_1   GUI_ID_USER + 3
#define MENU_ID_MIC_2   GUI_ID_USER + 4

/*********************************************************************
*
*       Local types
*
**********************************************************************
*/
typedef struct {
	OS_TASK guiTaskId;
        OS_STACKPTR int guiTaskStack[2048];
	OS_TASK fftTaskId;
        OS_STACKPTR int fftTaskStack[2048];
        OS_TASK tsTaskId;
        OS_STACKPTR int tsTaskStack[2048];

	GRAPH_Handle fftGraph;
	GRAPH_DATA_Handle fftGraphData;
	GRAPH_SCALE_Handle fftGraphScale;

        TEXT_Handle titleText;

	int16_t dmaBuffer[DMA_BUFFER_LENGTH];
	float32_t fftInput[SIGNAL_SAMPLES];
	float32_t fftOutput[SIGNAL_SAMPLES];
	int16_t fftDisplay[FFT_LENGTH];
} AppGlobals_s;

// Application global data and configuration.
static AppGlobals_s appGlobals;

// This structure is defined in the stm32746g_discovery_audio.c file and used in the IRQ handler.
extern SAI_HandleTypeDef haudio_in_sai;

/*********************************************************************
*
*       processMenuData()
*
*  Function description
*    Process the gesture messages.
*/

static void processGestureData(WM_MESSAGE *pMsg) {
  // Keep the information about last factor do distinguish between zoom in and zoom out evets.
  static int lastZoomingFactor;

  WM_GESTURE_INFO *info = (WM_GESTURE_INFO *)pMsg->Data.p;
  if ((info->Flags & WM_GF_PAN) == WM_GF_PAN) {
    if ((info->Point.x) > 0)
      OS_SignalEvent(TASK_EVENT_CHANGE_VIEW_MOVE_RIGHT, &appGlobals.fftTaskId);
    else if ((info->Point.x) < 0)
      OS_SignalEvent(TASK_EVENT_CHANGE_VIEW_MOVE_LEFT, &appGlobals.fftTaskId);
  } else if ((info->Flags & (WM_GF_ZOOM | WM_GF_BEGIN)) == (WM_GF_ZOOM | WM_GF_BEGIN)) {
    // Set the arbitrary initial factor for zoom event.
    info->Factor = 100;
    lastZoomingFactor = 100;
  } else if (((info->Flags & (WM_GF_ZOOM)) == WM_GF_ZOOM) && (info->Factor != 0)) {
    // The arbitrary factor change threshold for more gesture inertia.
    if ((lastZoomingFactor - info->Factor) > 25) {
      OS_SignalEvent(TASK_EVENT_CHANGE_VIEW_ZOOM_OUT_X, &appGlobals.fftTaskId);
      lastZoomingFactor = info->Factor;
    }
    if ((lastZoomingFactor - info->Factor) < -25) {
      OS_SignalEvent(TASK_EVENT_CHANGE_VIEW_ZOOM_IN_X, &appGlobals.fftTaskId);
      lastZoomingFactor = info->Factor;
    }
  }
}


/*********************************************************************
*
*       processMenuData()
*
*  Function description
*    Process the menu messages.
*/
static void processMenuData(WM_MESSAGE *pMsg) {
  MENU_MSG_DATA *pData = (MENU_MSG_DATA *)pMsg->Data.p;
  if (pData->MsgType == MENU_ON_ITEMPRESSED && pMsg->hWin == appGlobals.fftGraph) {
    if (pData->ItemId == MENU_ID_LINE_1)
      OS_SignalEvent(TASK_EVENT_CHANGE_SOURCE_LINE_1, &appGlobals.fftTaskId);
    if (pData->ItemId == MENU_ID_LINE_2)
      OS_SignalEvent(TASK_EVENT_CHANGE_SOURCE_LINE_2, &appGlobals.fftTaskId);
    if (pData->ItemId == MENU_ID_MIC_1)
      OS_SignalEvent(TASK_EVENT_CHANGE_SOURCE_MIC_1, &appGlobals.fftTaskId);
    if (pData->ItemId == MENU_ID_MIC_2)
      OS_SignalEvent(TASK_EVENT_CHANGE_SOURCE_MIC_2, &appGlobals.fftTaskId);
  }
}


/*********************************************************************
*
*       graphCallback()
*
*  Function description
*    Process the messages coming to the GRAPH window.
*/
static void graphCallback(WM_MESSAGE *pMsg) {
  if (pMsg->MsgId == WM_GESTURE) {

    if (pMsg->hWin != appGlobals.fftGraph)
      return;

    processGestureData(pMsg);

  } else {

    if (pMsg->MsgId == WM_MENU)
      processMenuData(pMsg);

    GRAPH_Callback(pMsg);
  }
}


/*********************************************************************
*
*       addMenuItem()
*
*/
static void addMenuItem(MENU_Handle menu, MENU_Handle submenu, char *name, uint16_t id) {
  MENU_ITEM_DATA item = {0};
  item.pText = name;
  item.hSubmenu = submenu;
  item.Id = id;
  MENU_AddItem(menu, &item);
}


/*********************************************************************
*
*       guiTask()
*
*  Function description
*    Both GRAPHs have also event callbacks for processing gesture data.
*/
static void guiTask(void) {

  WM_SetCreateFlags(WM_CF_MEMDEV | WM_CF_MEMDEV_ON_REDRAW);

  GUI_Init();

  GUI_MTOUCH_Enable(1);
  WM_GESTURE_Enable(1);

  GUI_Clear();

  appGlobals.fftGraph = GRAPH_CreateEx(0, 0, LCD_GetXSize(), LCD_GetYSize(), WM_HBKWIN, WM_CF_SHOW | WM_CF_GESTURE, 0, GUI_ID_GRAPH1);
  GRAPH_SetBorder(appGlobals.fftGraph, 5, 20, 5, 5);

  appGlobals.fftGraphScale = GRAPH_SCALE_Create(LCD_GetYSize() - 12, GUI_TA_CENTER, GRAPH_SCALE_CF_HORIZONTAL, 50);
  GRAPH_SCALE_SetFactor(appGlobals.fftGraphScale, ((1.0 * DEFAULT_AUDIO_IN_FREQ / SIGNAL_SAMPLES) / 2.0));
  GRAPH_AttachScale(appGlobals.fftGraph, appGlobals.fftGraphScale);

  appGlobals.fftGraphData = GRAPH_DATA_YT_Create(GUI_GREEN, FFT_LENGTH, NULL, 0);
  GRAPH_DATA_YT_SetAlign(appGlobals.fftGraphData, GRAPH_ALIGN_LEFT);
  GRAPH_AttachData(appGlobals.fftGraph, appGlobals.fftGraphData);

  WM_SetCallback(appGlobals.fftGraph, graphCallback);

  MENU_Handle hMenu = MENU_CreateEx(1, 1, 0, 0, appGlobals.fftGraph, 0, MENU_CF_HORIZONTAL, 0);
  MENU_Handle hMenuSource = MENU_CreateEx(0, 0, 0, 0, appGlobals.fftGraph, 0, MENU_CF_VERTICAL, 0);

  addMenuItem(hMenuSource, 0, "Line Ch R", MENU_ID_LINE_2);
  addMenuItem(hMenuSource, 0, "Line Ch L", MENU_ID_LINE_1);
  addMenuItem(hMenuSource, 0, "Mic R", MENU_ID_MIC_1);
  addMenuItem(hMenuSource, 0, "Mic L", MENU_ID_MIC_2);
  addMenuItem(hMenu, hMenuSource, "Source", 0);

  appGlobals.titleText = TEXT_CreateEx(100, 2, 0, 0, appGlobals.fftGraph, WM_CF_SHOW, TEXT_CF_LEFT, 0, "Select Input Source");

  BSP_TS_Init(LCD_GetXSize(), LCD_GetYSize());
  BSP_TS_ITConfig();

  while (1) {
    // Repaint the windows every 10 ms.
    GUI_Delay(10);
  }
}


/*********************************************************************
*
*       scaleAxisXFloat()
*
*  Function description
*    Scale the tab array values by calculating the mean value for 
*    samples number equal to scale parameter.
*/
static void scaleAxisXFloat(float32_t *tab, uint32_t len, uint32_t scale) {
  uint32_t index;
  for (index = 0; index < len / scale; index++)
    arm_mean_f32(tab + (index * scale), scale, tab + index);
}


/*********************************************************************
*
*       scaleAxisYFloat()
*
*  Function description
*    Scale the tab array values to fit the window size.
*/
static void scaleAxisYFloat(float32_t *tab, uint32_t len) {
  float32_t min;
  float32_t max;
  uint32_t index;

  min = 0;
  max = 2000000;

  for (index = 0; index < len; index++)
    tab[index] = (tab[index] - min) * GRAPH_RANGE_Y / (max - min) + GRAPH_OFFSET_Y;
}


/*********************************************************************
*
*       startRecording()
*
*  Function description
*    Initialize and start audio recording from given device.
*/
static void startRecording(uint8_t isRunning, uint16_t inputDevice) {
  OS_EnterRegion();
  if (isRunning > 0)
    BSP_AUDIO_IN_Stop(CODEC_PDWN_SW);
  BSP_AUDIO_IN_InitEx(inputDevice, DEFAULT_AUDIO_IN_FREQ, 16, 0);
  BSP_AUDIO_IN_Record((uint16_t *)appGlobals.dmaBuffer, DMA_BUFFER_LENGTH);
  OS_LeaveRegion();
}


/*********************************************************************
*
*       fftTask()
*
*  Function description
*    Task responsible for displaying the calculated fft on the screen.
*    Fft plot is updated after half and full DMA transfers.
*/
static void fftTask(void) {
  arm_rfft_fast_instance_f32 fftInit;
  arm_rfft_fast_init_f32(&fftInit, SIGNAL_SAMPLES);

  float32_t displayOffsetX = 0; // Offset of the plot as the samples number.
  float32_t displayScaleX = 1;  // Plot scaling factor.
  uint8_t channel = 0;          // Input signal channel.
  uint8_t running = 0;          // Recording running flag.

  uint32_t event;
  uint16_t idx;

  // Scale factor used for proper X axis scale display. It is divided by 2 because
  // only the real part of the fft is calculated.
  float32_t scaleFactor = (1.0 * DEFAULT_AUDIO_IN_FREQ / SIGNAL_SAMPLES) / 2.0;
  for (;;) {

    event = OS_WaitEvent(0xFFFFFFFF);
    if (event & TASK_EVENT_DMA_HALF_DONE) {
      for (idx = 0; idx < SIGNAL_SAMPLES; idx++)
        appGlobals.fftInput[idx] = appGlobals.dmaBuffer[idx * 2 + channel];
      arm_rfft_fast_f32(&fftInit, appGlobals.fftInput, appGlobals.fftOutput, 0);
      arm_abs_f32(appGlobals.fftOutput, appGlobals.fftOutput, SIGNAL_SAMPLES);

      scaleAxisXFloat(appGlobals.fftOutput, SIGNAL_SAMPLES, displayScaleX);
      scaleAxisYFloat(appGlobals.fftOutput, SIGNAL_SAMPLES / displayScaleX);
      for (idx = 0; idx < FFT_LENGTH; idx++) {
        appGlobals.fftDisplay[idx] = appGlobals.fftOutput[idx + (uint16_t)(displayOffsetX / displayScaleX)];
      }

      GRAPH_DetachData(appGlobals.fftGraph, appGlobals.fftGraphData);
      GRAPH_DATA_YT_Delete(appGlobals.fftGraphData);
      appGlobals.fftGraphData = GRAPH_DATA_YT_Create(GUI_GREEN, FFT_LENGTH, appGlobals.fftDisplay, FFT_LENGTH);
      GRAPH_AttachData(appGlobals.fftGraph, appGlobals.fftGraphData);
    }
    if (event & TASK_EVENT_DMA_DONE) {
      for (idx = 0; idx < SIGNAL_SAMPLES; idx++)
        appGlobals.fftInput[idx] = appGlobals.dmaBuffer[DMA_BUFFER_LENGTH / 2 + idx * 2 + channel];
      arm_rfft_fast_f32(&fftInit, appGlobals.fftInput, appGlobals.fftOutput, 0);
      arm_abs_f32(appGlobals.fftOutput, appGlobals.fftOutput, SIGNAL_SAMPLES);

      scaleAxisXFloat(appGlobals.fftOutput, SIGNAL_SAMPLES, displayScaleX);
      scaleAxisYFloat(appGlobals.fftOutput, SIGNAL_SAMPLES / displayScaleX);
      for (idx = 0; idx < FFT_LENGTH; idx++) {
        appGlobals.fftDisplay[idx] = appGlobals.fftOutput[idx + (uint16_t)(displayOffsetX / displayScaleX)];
      }

      GRAPH_DetachData(appGlobals.fftGraph, appGlobals.fftGraphData);
      GRAPH_DATA_YT_Delete(appGlobals.fftGraphData);
      appGlobals.fftGraphData = GRAPH_DATA_YT_Create(GUI_GREEN, FFT_LENGTH, appGlobals.fftDisplay, FFT_LENGTH);
      GRAPH_AttachData(appGlobals.fftGraph, appGlobals.fftGraphData);
    }
    if (event & TASK_EVENT_CHANGE_VIEW_MOVE_LEFT) {
      if (displayOffsetX + displayScaleX * (FFT_LENGTH + 10) < SIGNAL_SAMPLES) {
        displayOffsetX += 10 * displayScaleX;
        GRAPH_SCALE_SetOff(appGlobals.fftGraphScale, -1 * displayOffsetX / displayScaleX);
      }
    }
    if (event & TASK_EVENT_CHANGE_VIEW_MOVE_RIGHT) {
      if (displayOffsetX >= 10 * displayScaleX) {
        displayOffsetX -= 10 * displayScaleX;
        GRAPH_SCALE_SetOff(appGlobals.fftGraphScale, -1 * displayOffsetX / displayScaleX);
      }
    }
    if (event & TASK_EVENT_CHANGE_VIEW_ZOOM_IN_X) {
      if (displayScaleX > 1) {
        displayOffsetX *= scaleFactor;
        displayOffsetX += FFT_LENGTH * displayScaleX / 4;
        displayScaleX--;
        displayOffsetX /= scaleFactor;

        GRAPH_SCALE_SetFactor(appGlobals.fftGraphScale, scaleFactor * displayScaleX);
        GRAPH_SCALE_SetOff(appGlobals.fftGraphScale, -1 * (displayOffsetX / displayScaleX));
      }
    }
    if (event & TASK_EVENT_CHANGE_VIEW_ZOOM_OUT_X) {
      if ((FFT_LENGTH - 1) * (displayScaleX + 1) < SIGNAL_SAMPLES) {
        displayOffsetX *= scaleFactor;
        displayOffsetX -= FFT_LENGTH * displayScaleX / 2;
        displayScaleX++;
        displayOffsetX /= scaleFactor;

        if (displayOffsetX < 0)
          displayOffsetX = 0;
        else if (displayScaleX * (displayOffsetX + FFT_LENGTH) >= SIGNAL_SAMPLES)
          displayOffsetX = SIGNAL_SAMPLES / displayScaleX - FFT_LENGTH;
        GRAPH_SCALE_SetFactor(appGlobals.fftGraphScale, scaleFactor * displayScaleX);
        GRAPH_SCALE_SetOff(appGlobals.fftGraphScale, -1 * (uint16_t)(displayOffsetX / displayScaleX));
      }
    }
    if (event & TASK_EVENT_CHANGE_SOURCE_LINE_1) {
      startRecording(running, INPUT_DEVICE_INPUT_LINE_1);
      channel = 0;
      running = 1;
      TEXT_SetText(appGlobals.titleText, "Line Input Left Channel");
    }
    if (event & TASK_EVENT_CHANGE_SOURCE_LINE_2) {
      startRecording(running, INPUT_DEVICE_INPUT_LINE_1);
      channel = 1;
      running = 1;
      TEXT_SetText(appGlobals.titleText, "Line Input Right Channel");
    }
    if (event & TASK_EVENT_CHANGE_SOURCE_MIC_1) {
      startRecording(running, INPUT_DEVICE_DIGITAL_MICROPHONE_2);
      channel = 0;
      running = 1;
      TEXT_SetText(appGlobals.titleText, "Microphone Right");
    }
    if (event & TASK_EVENT_CHANGE_SOURCE_MIC_2) {
      startRecording(running, INPUT_DEVICE_DIGITAL_MICROPHONE_2);
      channel = 1;
      running = 1;
      TEXT_SetText(appGlobals.titleText, "Microphone Left");
    }
  }
}


/*********************************************************************
*
*       storeMtouchEvent()
*
*/
static void storeMtouchEvent(TS_StateTypeDef *tsState, int eventType) {
  GUI_MTOUCH_EVENT event;
  event.LayerIndex = 0;
  event.TimeStamp = OS_GetTime();

  GUI_MTOUCH_INPUT input[2];

  if (eventType == GUI_MTOUCH_FLAG_UP) {
    event.NumPoints = 2;
    input[0].Flags = GUI_MTOUCH_FLAG_UP;
    input[1].Flags = GUI_MTOUCH_FLAG_UP;
  } else if (GUI_MTOUCH_FLAG_DOWN || GUI_MTOUCH_FLAG_MOVE) {
    event.NumPoints = tsState->touchDetected;
    input[0].Id = 0;
    input[0].x = tsState->touchX[0];
    input[0].y = tsState->touchY[0];
    input[0].Flags = eventType;
    input[1].Id = 0;
    input[1].x = tsState->touchX[1];
    input[1].y = tsState->touchY[1];
    input[1].Flags = eventType;
  }

  GUI_MTOUCH_StoreEvent(&event, input);
}


/*********************************************************************
*
*       storePidEvent()
*
*/
static void storePidEvent(TS_StateTypeDef *tsState, uint8_t pressed) {
  GUI_PID_STATE pidState;
  pidState.Layer = 0;

  if (pressed == 0) {
    pidState.Pressed = 0;
    pidState.x = -1;
    pidState.y = -1;
  } else {
    pidState.Pressed = 1;
    pidState.x = tsState->touchX[0];
    pidState.y = tsState->touchY[0];
  }
  GUI_PID_StoreState(&pidState);
}


/*********************************************************************
*
*       tsTask()
*
*  Function description
*    Receive the data from the touch screen after getting event from
*    the interrupt.
*/
static void tsTask(void) {
  OS_TASK_EVENT event;
  int moving = 0;
  int pressing = 0;
  int lastPoints = 0;

  while (1) {
    event = OS_WaitEvent_Timed(0xFFFFFFFF, 100);
    if (event == 0 && moving > 0) {
      moving = 0;
      pressing = 0;
      storePidEvent(NULL, pressing);
      storeMtouchEvent(NULL, GUI_MTOUCH_FLAG_UP);
    } else if (event != 0) {
      TS_StateTypeDef state;

      OS_EnterRegion();
      uint8_t retCode = BSP_TS_GetState(&state);
      OS_LeaveRegion();

      if (retCode != TS_OK)
        continue;

      if (state.touchDetected < 1 || state.touchDetected > 2)
        continue;

      if (lastPoints != state.touchDetected) {
        moving = 0;
        lastPoints = state.touchDetected;
        storeMtouchEvent(NULL, GUI_MTOUCH_FLAG_UP);
      }

      if (moving > 0) {
        storeMtouchEvent(&state, GUI_MTOUCH_FLAG_MOVE);
      } else {
        moving = 1;
        storeMtouchEvent(&state, GUI_MTOUCH_FLAG_DOWN);
      }

      if (pressing == 0 && state.touchDetected == 1) {
        pressing = 1;

        storePidEvent(&state, pressing);
      }

      OS_Delay(10);
    }
  }
}


/*********************************************************************
*
*       main()
*/
int main(void) {
  OS_InitKern();
  OS_InitHW();
  OS_CREATETASK(&appGlobals.guiTaskId, "GuiTask", guiTask, 150, appGlobals.guiTaskStack);
  OS_CREATETASK(&appGlobals.fftTaskId, "FftTask", fftTask, 150, appGlobals.fftTaskStack);
  OS_CREATETASK(&appGlobals.tsTaskId, "tsTask", tsTask, 150, appGlobals.tsTaskStack);
  OS_Start();
  return 0;
}

// Interrupt from the DMA engine.
void AUDIO_IN_SAIx_DMAx_IRQHandler(void) {
  OS_INT_Enter();
  HAL_DMA_IRQHandler(haudio_in_sai.hdmarx);
  OS_INT_Leave();
}

// Interrupt from the touch screen.
void EXTI15_10_IRQHandler(void) {
  OS_INT_Enter();
  HAL_GPIO_EXTI_IRQHandler(TS_INT_PIN);
  OS_INT_Leave();
}

// This callback nofities the tsTask about touch screen event.
void HAL_GPIO_EXTI_Callback(uint16_t GPIO_Pin) {
  (void)GPIO_Pin;
  OS_SignalEvent(TASK_EVENT_NEW_GESTURE_DATA, &appGlobals.tsTaskId);
}

// This callback is called when the half of the data is received. It notifies
// the fft task about data ready for processing.
void BSP_AUDIO_IN_HalfTransfer_CallBack(void) {
  OS_SignalEvent(TASK_EVENT_DMA_HALF_DONE, &appGlobals.fftTaskId);
}

// This callback is called when the half of the data is received. It notifies
// the fft task about data ready for processing.
void BSP_AUDIO_IN_TransferComplete_CallBack(void) {
  OS_SignalEvent(TASK_EVENT_DMA_DONE, &appGlobals.fftTaskId);
}

/*************************** End of file ****************************/